/*
 * ################################################################
 *
 * ProActive Parallel Suite(TM): The Java(TM) library for
 *    Parallel, Distributed, Multi-Core Computing for
 *    Enterprise Grids & Clouds
 *
 * Copyright (C) 1997-2012 INRIA/University of
 *                 Nice-Sophia Antipolis/ActiveEon
 * Contact: proactive@ow2.org or contact@activeeon.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation; version 3 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * If needed, contact us to obtain a release under GPL Version 2 or 3
 * or a different license than the AGPL.
 *
 *  Initial developer(s):               The ProActive Team
 *                        http://proactive.inria.fr/team_members.htm
 *  Contributor(s):
 *
 * ################################################################
 */
package org.objectweb.proactive.mavenplugin;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URLClassLoader;
import java.util.List;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;


/**
 * This class is used to provide the operations which are common to all the
 * ProActive class generations.
 * 
 * @phase compile
 * @requiresDependencyResolution compile+runtime
 * 
 * @author The ProActive Team
 */
public abstract class AbstractClassGeneratorMojo extends AbstractMojo {

    /**
     * Compile classpath of the maven project.
     * 
     * @parameter expression="${project.compileClasspathElements}"
     * @readonly
     */
    protected List<String> projectClasspathElements;

    /**
     * Specifies where to place the generated class files.
     * 
     * @parameter default-value="${project.build.outputDirectory}"
     */
    protected File outputDirectory;

    protected URLClassLoader classLoader;

    /**
     * {@inheritDoc}
     */
    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        this.classLoader = Util.createClassLoader(this.projectClasspathElements);

        this.init();

        List<String> classNames = this.getClassNames();
        Thread.currentThread().setContextClassLoader(this.classLoader);
        for (final String className : classNames) {
            try {
                this.classLoader.loadClass(className);
            } catch (ClassNotFoundException cnfe) {
                try {
                    byte[] data = this.generateClass(className);

                    String fileName = this.writeClass(this.outputDirectory, className, data);

                    this.getLog().info("Generated " + this.getKind() + " " + fileName);
                } catch (Exception e) {
                    this.getLog().error("Failed to generate " + this.getKind() + " " + className, e);
                }
            }
        }
    }

    protected abstract void init() throws MojoExecutionException;

    protected abstract List<String> getClassNames() throws MojoExecutionException;

    protected abstract byte[] generateClass(String className) throws Exception;

    public String writeClass(File outputDirectory, String className, byte[] data) throws IOException {
        // writes the bytecode into a file
        String fileName = new File(outputDirectory.toString(), className.replace('.', File.separatorChar) +
            ".class").toString();
        FileOutputStream fos = null;

        try {
            new File(fileName.substring(0, fileName.lastIndexOf(File.separatorChar))).mkdirs();

            // dumps the bytecode into the file
            fos = new FileOutputStream(new File(fileName));
            fos.write(data);
            fos.flush();
        } finally {
            try {
                fos.close();
            } catch (IOException ioe) {
                this.getLog().error("Failed to close " + fileName, ioe);
            }
        }

        return fileName;
    }

    protected abstract String getKind();

}
